/* Copyright 2006 by Daniel Bump                                     *
 *                                                                   *
 * This program is free software; you can redistribute it and/or     *
 * modify it under the terms of the GNU General Public License as    *
 * published by the Free Software Foundation - version 2             *
 */

#include <stdio.h>
#include <unistd.h>
#include "drawrubik.h"

/*

The purpose of this program is to create metapost images of
Rubik's cubes.

Input consists of lines of three types.

COMMENT LINE:

An empty line, or any line beginning with whitespace is a 
comment line, and is ignored.

METAPOST LINE:

[filename] followed by a space, then a word, which is a
sequence of letters from the alphabet:

U D R L F B ' ^ 2 m ( )

Output is a metapost file [filename].mp which draws the configuration
of the cube after the indicated sequence of moves. Multiple lines of
this type will create as many different metapost files.

COMMAND LINE:

A line that begins with +,-,* consists of a series of commands
that set the values of some global modifiers. + turns the
modifier on, - turns it off. 

The global modifiers affected by + and - are:

s   stereo         stereographic image
b   bottom         show bottom of cube
l   label          print label below diagram
x   initialize     return the cube to unscrambled position

The modification is persistent, that is, it continues until it is
countermanded by another such command.

Exception: `initialize` is not persistent, but reverts to the 
default value of true after one diagram is printed. Preceding
a diagram by a command line containing -x means that the cube
operations are continued from the previous diagram, rather
from an unscrambled cube.

After * one may give a "mask" consisting of a sequence
of faces that are to be darkened. Unlike the + and - 
commands, this modification is not persistent but only
affects the next metapost file. Thus to darken the
top corner pieces:

* u0 u2 u6 u7 b8 r2 r8 f2 f8 l2 l8 b2

Here is the labelling of the cube in stereographic projection:


D6                  D7                  D8


             B0     B3     B6
               B1   B4   B7
       L6        B2 B5 B8        R0
          L7                  R1
             L8  U2 U1 U0  R2
D3     L3 L4 L5  U5 U4 U3  R5 R4 R3     D5
             L2  U8 U7 U6  R8
          L1                  R7
       L0        F8 F5 F2        R6
               F7   F4   F1
             F6     F3     F0


D0                  D1                  D2

Input is terminated by an empty line or end of file.

*/

int up[9], right[9], front[9], left[9], down[9], back[9];
int upmask[9], rightmask[9], frontmask[9];
int leftmask[9], downmask[9], backmask[9];
int stereo = 0;
int bottom = 0;
int labelfigure = 0;
int initialize = 1;
int debug = 0;

static const char *color_names[] = {
  COLOR_NAMES
};

int
main (int argc, char **argv)
{
  char nextchar;
  int finished = 0;
  int c;

  initialize_mask();
  while ((c = getopt (argc, argv, "dh")) != -1)
    switch (c) 
      {
      case 'd' :
	debug = 1;
	break;
      case 'h' :
	printf("%s", HELP);
	return;
      }
    while (!finished) {
    if ((c = getchar()) == EOF) {
      finished = 1;
      break;
    }
    else ungetc(c, stdin);

    if ((c == '+') || (c == '-'))
      command_line();
    else if (c == '*')
      darken_line();
    else if ((c == ' ') || (c == '\t') || (c == '\n'))
      comment_line();
    else metapost_line();
  }
  return 0;
}

void
initialize_cube()
{
  int i;
  
  for (i = 0; i < 9; i++) {
    up[i] = BLUE;
    right[i] = RED;
    front[i] = WHITE;
    left[i] = ORANGE;
    down[i] = GREEN;
    back[i] = YELLOW;
  }
}

/* processes a comand line */

void
command_line()
{
  int c;
  int toggleon = 0;

  while ((c = getchar()) != '\n') {
    if (c == '+')
      toggleon = 1;
    else if (c == '-')
      toggleon = 0;
    else if (toggleon) {
      if (c == 'l')
	labelfigure = 1;
      else if (c == 's')
	stereo = 1;
      else if (c == 'b')
	bottom = 1;
      else if (c == 'x')
	initialize = 1;
    }
    else {
      if (c == 'l')
	labelfigure = 0;
      else if (c == 's')
	stereo = 0;
      else if (c == 'b')
	bottom = 0;
      else if (c == 'x')
	initialize = 0;
    }
  }
}

void
initialize_mask()
{
  int i;

  for (i = 0; i < 9; i++) {
    upmask[i] = 0;
    downmask[i] = 0;
    frontmask[i] = 0;
    backmask[i] = 0;
    rightmask[i] = 0;
    leftmask[i] = 0;
  }
}


void
darken_line()
{
  int c;
  int n;
  int toggleon = 0;

  while ((c = getchar()) != '\n') {
    if (c == 'd') {
      scanf("%d", &n);
      downmask[n] = 1;
    }
    if (c == 'u') {
      scanf("%d", &n);
      upmask[n] = 1;
    }
    if (c == 'f') {
      scanf("%d", &n);
      frontmask[n] = 1;
    }
    if (c == 'b') {
      scanf("%d", &n);
      backmask[n] = 1;
    }
    if (c == 'r') {
      scanf("%d", &n);
      rightmask[n] = 1;
    }
    if (c == 'l') {
      scanf("%d", &n);
      leftmask[n] = 1;
    }
  }
}


int
comment_line()
{
  int c;

  while ((c = getchar()) != '\n');
}

/* processes a line to make a metapost file */

int
metapost_line()
{
  FILE *mpfile;
  int j = 0;
  int k;
  char token, slice, lasttoken, lastslice;
  char filename[80];
  char label[MAXWORD];
  char parsed[MAXWORD];
  int paren = 0;

  memset(label, 0, sizeof(label));
  memset(parsed, 0, sizeof(parsed));

  if (initialize) {
    if (debug)
      printf("initializing cube\n");
    initialize_cube();
  }
  else 
    initialize = 1;
  if (!scanf("%s[ \n]", filename)) {
    fprintf(stderr, "Usage: outputfilename word\n");
    return 0;
  }
  strcat(filename + strlen(filename), ".mp");
  mpfile = fopen(filename, "w");
  fprintf(mpfile, PREAMBLE);
  while (((token = getchar()) != EOF) && token != '\n')
    if (token != ' ')
      label[j++] = token;
  label[j] = 0;
  parse_word(label, parsed);

  if (debug)
    printf("label: %s parsed: %s\n", label, parsed);
  
  j = 0;
  while ((token = parsed[j]) != 0) {
    slice = 0;
    
    if ((parsed[j+1] == 'm') || 
	(parsed[j+1] == '_') && (parsed[j+2] == 'm'))
      slice = 1;

    if ((token == 'R') || (token == 'L') || (token == 'U') ||
	(token == 'D') || (token == 'F') || (token == 'B')) {
      process_token(token, mpfile, slice);
      lastslice = slice;
      lasttoken = token;
    }
    else if (token == '\'') {
	process_token(lasttoken, mpfile, lastslice);
	process_token(lasttoken, mpfile, lastslice);
      }
    else if (token == '2')
      process_token(lasttoken, mpfile, lastslice);
    j++;
  }
  {
    int i;
    for (i = 0; i < 9; i++) {
      if (downmask[i])
	down[i] = DARK;
      if (upmask[i])
	up[i] = DARK;
      if (leftmask[i])
	left[i] = DARK;
      if (rightmask[i])
	right[i] = DARK;
      if (frontmask[i])
	front[i] = DARK;
      if (backmask[i])
	back[i] = DARK;
    }
  }
  if (stereo) 
    fprintf(mpfile, STEREO,
	    color_names[right[0]],
	    color_names[right[1]],
	    color_names[right[2]],
	    color_names[right[3]],
	    color_names[right[4]],
	    color_names[right[5]],
	    color_names[right[6]],
	    color_names[right[7]],
	    color_names[right[8]],
	    color_names[up[0]],
	    color_names[up[1]],
	    color_names[up[2]],
	    color_names[up[3]],
	    color_names[up[4]],
	    color_names[up[5]],
	    color_names[up[6]],
	    color_names[up[7]],
	    color_names[up[8]],
	    color_names[front[0]],
	    color_names[front[1]],
	    color_names[front[2]],
	    color_names[front[3]],
	    color_names[front[4]],
	    color_names[front[5]],
	    color_names[front[6]],
	    color_names[front[7]],
	    color_names[front[8]],
	    color_names[back[0]],
	    color_names[back[1]],
	    color_names[back[2]],
	    color_names[back[3]],
	    color_names[back[4]],
	    color_names[back[5]],
	    color_names[back[6]],
	    color_names[back[7]],
	    color_names[back[8]],
	    color_names[left[0]],
	    color_names[left[1]],
	    color_names[left[2]],
	    color_names[left[3]],
	    color_names[left[4]],
	    color_names[left[5]],
	    color_names[left[6]],
	    color_names[left[7]],
	    color_names[left[8]],
	    label);
  else if (bottom)
    fprintf(mpfile, TOPLEFT,
	    color_names[back[8]],
	    color_names[back[7]],
	    color_names[back[6]],
	    color_names[back[5]],
	    color_names[back[4]],
	    color_names[back[3]],
	    color_names[back[2]],
	    color_names[back[1]],
	    color_names[back[0]],
	    color_names[down[8]],
	    color_names[down[5]],
	    color_names[down[2]],
	    color_names[down[7]],
	    color_names[down[4]],
	    color_names[down[1]],
	    color_names[down[6]],
	    color_names[down[3]],
	    color_names[down[0]],
	    color_names[left[8]],
	    color_names[left[7]],
	    color_names[left[6]],
	    color_names[left[5]],
	    color_names[left[4]],
	    color_names[left[3]],
	    color_names[left[2]],
	    color_names[left[1]],
	    color_names[left[0]]);
  else
    fprintf(mpfile, TOPLEFT,
	    color_names[right[0]],
	    color_names[right[1]],
	    color_names[right[2]],
	    color_names[right[3]],
	    color_names[right[4]],
	    color_names[right[5]],
	    color_names[right[6]],
	    color_names[right[7]],
	    color_names[right[8]],
	    color_names[up[0]],
	    color_names[up[1]],
	    color_names[up[2]],
	    color_names[up[3]],
	    color_names[up[4]],
	    color_names[up[5]],
	    color_names[up[6]],
	    color_names[up[7]],
	    color_names[up[8]],
	    color_names[front[0]],
	    color_names[front[1]],
	    color_names[front[2]],
	    color_names[front[3]],
	    color_names[front[4]],
	    color_names[front[5]],
	    color_names[front[6]],
	    color_names[front[7]],
	    color_names[front[8]]);
  if (labelfigure)
    fprintf(mpfile, "label(btex $%s$ etex, z400);\n\n", label);
  fprintf(mpfile, EPILOG);
  fclose(mpfile);
  initialize = 1;
  initialize_mask();
}

int
parse_word(char *label, char *parsed)
{
  char preparsed[MAXWORD];

  int done = 0;
  int exponent = 1;

  strcpy(preparsed, label);

  /* copy the string in preparsed to parsed. If a
     left paren is encountered, copy the block inside
     multiple times (depending on the exponent) while
     removing the matching parentheses. If this happens,
     repeat no left parentheses is found. */
  
  while (!done) {
    int j = 0;
    int k;

    if (debug)
      printf("preparsed: %s\n", preparsed);
    done = 1; /* no left paren yet seen */
    while (preparsed[j] != 0) {
      parsed[j] = preparsed[j];
      if (preparsed[j] == '(') {
	/* Find matching right parenthesis */
	int parencount = 1;
	int exponent = 1;
	int l;
	
	k = j+1;
	done = 0;
	
	while (parencount > 0) {
	  if (!preparsed[k]) {
	    fprintf(stderr, "unmatched left parenthesis: %s\n", preparsed);
	    return 0;
	  }
	  if (preparsed[k] == '(')
	    parencount++;
	  else if (preparsed[k] == ')')
	    parencount--;
	  if (parencount != 0)
	    k++;
	}
	/* now j = location of left paren, k = location of right paren */
	if (preparsed[k+1] == '^')
	  /* ASCII value of 0,1,2, ... is 48,49,50, ... */
	  exponent = preparsed[k+2] - 48;
	/* copy exponent times the block in paren */
	for (l = 0; l < exponent; l++)
	  strncpy(parsed+j+l*(k-j-1), preparsed+j+1, k-j-1);
	/* copy the rest of the line, omitting the exponent */
	if (preparsed[k+1] == '^') {
	  if (preparsed[k+3] == 0)
	    parsed[j+exponent*(k-j-1)] = 0;
	  else
	    strcpy(parsed+j+exponent*(k-j-1), preparsed+k+3);
	}
	else {
	  if (preparsed[k+1] == 0)
	    parsed[j+exponent*(k-j-1)] = 0;
	  else
	    strcpy(parsed+j+exponent*(k-j-1), preparsed+k+1);
	}
      }
      if (!done) {
	strcpy(preparsed, parsed);
	j = 0;
	done = 1;
      }
      else j++;
    }
  }
}

void
process_token(char token, FILE *outfile, int slice) {
  int oldup[9], oldright[9], oldfront[9], oldleft[9], olddown[9], oldback[9];
  int i;
  
  if (debug)
    printf("processing token %c (slice %d)\n", token, slice);

  for (i = 0; i < 9; i++) {
    oldup[i] = up[i];
    oldright[i] = right[i];
    oldfront[i] = front[i];
    oldleft[i] = left[i];
    olddown[i] = down[i];
    oldback[i] = back[i];
  }
  if (token == 'U') {
    if (slice) {
      left[7]=oldfront[7];
      left[4]=oldfront[4];
      left[1]=oldfront[1];
      front[7]=oldright[7];
      front[4]=oldright[4];
      front[1]=oldright[1];
      right[7]=oldback[7];
      right[4]=oldback[4];
      right[1]=oldback[1];
      back[7]=oldleft[7];
      back[4]=oldleft[4];
      back[1]=oldleft[1];
    } else {
      up[0] = oldup[2];
      up[1] = oldup[5];
      up[2] = oldup[8];
      up[3] = oldup[1];
      up[5] = oldup[7];
      up[6] = oldup[0];
      up[7] = oldup[3];
      up[8] = oldup[6];
      back[2] = oldleft[2];
      back[5] = oldleft[5];
      back[8] = oldleft[8];
      left[2] = oldfront[2];
      left[5] = oldfront[5];
      left[8] = oldfront[8];
      front[2] = oldright[2];
      front[5] = oldright[5];
      front[8] = oldright[8];
      right[2] = oldback[2];
      right[5] = oldback[5];
      right[8] = oldback[8];
    }
  }   
  else if (token == 'D') {
    if (slice) {
      front[7] = oldleft[7];
      front[4] = oldleft[4];
      front[1] = oldleft[1];
      right[7] = oldfront[7];
      right[4] = oldfront[4];
      right[1] = oldfront[1];
      back[7] = oldright[7];
      back[4] = oldright[4];
      back[1] = oldright[1];
      left[7] = oldback[7];
      left[4] = oldback[4];
      left[1] = oldback[1];
    } else {
      down[0] = olddown[6];
      down[1] = olddown[3];
      down[2] = olddown[0];
      down[3] = olddown[7];
      down[5] = olddown[1];
      down[6] = olddown[8];
      down[7] = olddown[5];
      down[8] = olddown[2];
      back[0] = oldright[0];
      back[3] = oldright[3];
      back[6] = oldright[6];
      right[0] = oldfront[0];
      right[3] = oldfront[3];
      right[6] = oldfront[6];
      front[0] = oldleft[0];
      front[3] = oldleft[3];
      front[6] = oldleft[6];
      left[0] = oldback[0];
      left[3] = oldback[3];
      left[6] = oldback[6];
    }
  }
  else if (token == 'L') {
    if (slice) {
      up[1] = oldback[3];
      up[4] = oldback[4];
      up[7] = oldback[5];
      front[3] = oldup[7];
      front[4] = oldup[4];
      front[5] = oldup[1];
      down[7] = oldfront[3];
      down[4] = oldfront[4];
      down[1] = oldfront[5];
      back[3] = olddown[1];
      back[4] = olddown[4];
      back[5] = olddown[7];
    } else {
      left[0] = oldleft[2];
      left[1] = oldleft[5];
      left[2] = oldleft[8];
      left[3] = oldleft[1];
      left[5] = oldleft[7];
      left[6] = oldleft[0];
      left[7] = oldleft[3];
      left[8] = oldleft[6];
      down[0] = oldfront[8];
      down[3] = oldfront[7];
      down[6] = oldfront[6];
      front[8] = oldup[2];
      front[7] = oldup[5];
      front[6] = oldup[8];
      up[2] = oldback[0];
      up[5] = oldback[1];
      up[8] = oldback[2];
      back[0] = olddown[0];
      back[1] = olddown[3];
      back[2] = olddown[6];
    }
  }
  else if (token == 'R') {
    if (slice) {
      back[3] = oldup[1];
      back[4] = oldup[4];
      back[5] = oldup[7];
      up[7] = oldfront[3];
      up[4] = oldfront[4];
      up[1] = oldfront[5];
      front[3] = olddown[7];
      front[4] = olddown[4];
      front[5] = olddown[1];
      down[1] = oldback[3];
      down[4] = oldback[4];
      down[7] = oldback[5];
    } else {
      right[0] = oldright[2];
      right[1] = oldright[5];
      right[2] = oldright[8];
      right[3] = oldright[1];
      right[5] = oldright[7];
      right[6] = oldright[0];
      right[7] = oldright[3];
      right[8] = oldright[6];
      down[8] = oldback[8];
      down[5] = oldback[7];
      down[2] = oldback[6];
      back[6] = oldup[0];
      back[7] = oldup[3];
      back[8] = oldup[6];
      up[0] = oldfront[2];
      up[3] = oldfront[1];
      up[6] = oldfront[0];
      front[0] = olddown[8];
      front[1] = olddown[5];
      front[2] = olddown[2];
    }
  }
  else if (token == 'F') {
    if (slice) {
      up[3] = oldleft[5];
      up[4] = oldleft[4];
      up[5] = oldleft[3];
      right[3] = oldup[3];
      right[4] = oldup[4];
      right[5] = oldup[5];
      down[3] = oldright[3];
      down[4] = oldright[4];
      down[5] = oldright[5];
      left[5] = olddown[3];
      left[4] = olddown[4];
      left[3] = olddown[5];
    } else {
      front[0] = oldfront[2];
      front[1] = oldfront[5];
      front[2] = oldfront[8];
      front[3] = oldfront[1];
      front[5] = oldfront[7];
      front[6] = oldfront[0];
      front[7] = oldfront[3];
      front[8] = oldfront[6];
      down[0] = oldright[6];
      down[1] = oldright[7];
      down[2] = oldright[8];
      right[6] = oldup[6];
      right[7] = oldup[7];
      right[8] = oldup[8];
      up[6] = oldleft[2];
      up[7] = oldleft[1];
      up[8] = oldleft[0];
      left[0] = olddown[2];
      left[1] = olddown[1];
      left[2] = olddown[0];
    }
  }
  else if (token == 'B') {
    if (slice) {
      left[5] = oldup[3];
      left[4] = oldup[4];
      left[3] = oldup[5];
      up[3] = oldright[3];
      up[4] = oldright[4];
      up[5] = oldright[5];
      right[3] = olddown[3];
      right[4] = olddown[4];
      right[5] = olddown[5];
      down[3] = oldleft[5];
      down[4] = oldleft[4];
      down[5] = oldleft[3];
    } else {
      back[0] = oldback[2];
      back[1] = oldback[5];
      back[2] = oldback[8];
      back[3] = oldback[1];
      back[5] = oldback[7];
      back[6] = oldback[0];
      back[7] = oldback[3];
      back[8] = oldback[6];
      down[6] = oldleft[8];
      down[7] = oldleft[7];
      down[8] = oldleft[6];
      left[6] = oldup[2];
      left[7] = oldup[1];
      left[8] = oldup[0];
      up[0] = oldright[0];
      up[1] = oldright[1];
      up[2] = oldright[2];
      right[0] = olddown[6];
      right[1] = olddown[7];
      right[2] = olddown[8];
    }
  }
}


/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */
